def create_moderation(create_moderation_request):  # noqa: E501
    """Classifies if text is potentially harmful.

     # noqa: E501

    :param create_moderation_request:
    :type create_moderation_request: dict | bytes

    :rtype: Union[CreateModerationResponse, Tuple[CreateModerationResponse, int], Tuple[CreateModerationResponse, int, Dict[str, str]]
    """
    raise NotImplementedError



import time
import uuid
from typing import List

from openai.pagination import SyncCursorPage
from openai.types.fine_tuning.fine_tuning_job import FineTuningJob, Hyperparameters
from openai.types.fine_tuning.fine_tuning_job_event import FineTuningJobEvent
from prefect.deployments import run_deployment
from prefect.deployments.flow_runs import FlowRun

# from timestep.database import InstanceStoreSingleton  # noqa: E501

# from timestep.worker import train_model

# instance_store = InstanceStoreSingleton()


def cancel_fine_tuning_job(fine_tuning_job_id):  # noqa: E501
    """Immediately cancel a fine-tune job.

     # noqa: E501

    :param fine_tuning_job_id: The ID of the fine-tuning job to cancel.
    :type fine_tuning_job_id: str

    :rtype: Union[FineTuningJob, Tuple[FineTuningJob, int], Tuple[FineTuningJob, int, Dict[str, str]]
    """
    raise NotImplementedError


# def create_fine_tuning_job(create_fine_tuning_job_request):  # noqa: E501
async def create_fine_tuning_job(body, token_info, user):
    """Creates a fine-tuning job which begins the process of creating a new model from a given dataset.  Response includes details of the enqueued job including job status and the name of the fine-tuned models once complete.  [Learn more about fine-tuning](/docs/guides/fine-tuning)

     # noqa: E501

    :param create_fine_tuning_job_request:
    :type create_fine_tuning_job_request: dict | bytes

    :rtype: Union[FineTuningJob, Tuple[FineTuningJob, int], Tuple[FineTuningJob, int, Dict[str, str]]
    """
    hyperparameters = Hyperparameters(
        # n_epochs="auto",
        n_epochs=1,
    )
    suffix = body.get("suffix")

    fine_tuning_job = FineTuningJob(
        id=str(uuid.uuid4()),
        created_at=int(time.time()),
        hyperparameters=hyperparameters,
        model=body.get("model"),
        object="fine_tuning.job",
        organization_id="organization_id",
        result_files=[],
        seed=body.get("seed", 42),
        status="queued",  # Literal['validating_files', 'queued', 'running', 'succeeded', 'failed', 'cancelled'],
        training_file=body.get("training_file"),
        validation_file=body.get("validation_file"),
    )

    # instance_store._shared_instance_state["fine_tuning_jobs"][fine_tuning_job.id] = fine_tuning_job
    instance_store.insert(fine_tuning_job)

    fine_tuning_job_id = fine_tuning_job.id

    fine_tuning_job = instance_store.select(FineTuningJob, id=fine_tuning_job_id)
    # assert uuid.UUID(fine_tuning_job.id, version=4) == uuid.UUID(fine_tuning_job_id, version=4)
    # assert uuid.UUID(fine_tuning_job.id) == uuid.UUID(fine_tuning_job_id)
    print("type(fine_tuning_job.id):", type(fine_tuning_job.id))
    assert str(fine_tuning_job.id) == fine_tuning_job_id

    fine_tuning_job_event = FineTuningJobEvent(
        id=str(uuid.uuid4()),
        created_at=int(time.time()),
        level="info",
        message="TODO: figure out what this messages should be...",
        object="fine_tuning.job.event",
    )

    instance_store._shared_instance_state["fine_tuning_job_events"][
        fine_tuning_job.id
    ] = [fine_tuning_job_event]

    # TODO: make this async, etc.
    # await train_model(fine_tuning_job_id=fine_tuning_job.id, suffix=suffix)
    flow_run: FlowRun = await run_deployment(
        idempotency_key=fine_tuning_job.id,
        name="agent-flow/agent-flow-deployment",
        # parameters={"names": ["Alice", "Bob"]},
        parameters={
            "inputs": {
                "fine_tuning_job_id": fine_tuning_job.id,
                "suffix": suffix,
            },
            "work_type": "train_agent",
        },
        # job_variables={"env": {"MY_ENV_VAR": "staging"}},
        timeout=0,  # don't wait for the run to finish
    )

    print("flow_run: ", flow_run)

    return fine_tuning_job.model_dump(mode="json")


# def list_fine_tuning_events(fine_tuning_job_id, after=None, limit=None):  # noqa: E501
def list_fine_tuning_events(fine_tuning_job_id, limit, token_info, user):
    """Get status updates for a fine-tuning job.

     # noqa: E501

    :param fine_tuning_job_id: The ID of the fine-tuning job to get events for.
    :type fine_tuning_job_id: str
    :param after: Identifier for the last event from the previous pagination request.
    :type after: str
    :param limit: Number of events to retrieve.
    :type limit: int

    :rtype: Union[ListFineTuningJobEventsResponse, Tuple[ListFineTuningJobEventsResponse, int], Tuple[ListFineTuningJobEventsResponse, int, Dict[str, str]]
    """
    # TODO: handle AsyncCursorPage?
    fine_tuning_job_events: List[FineTuningJobEvent] = (
        instance_store._shared_instance_state["fine_tuning_job_events"][
            fine_tuning_job_id
        ][-limit:]
    )

    sync_cursor_page: SyncCursorPage[FineTuningJobEvent] = SyncCursorPage(
        data=fine_tuning_job_events,
    )

    return sync_cursor_page.model_dump(mode="json")


def list_fine_tuning_job_checkpoints(
    fine_tuning_job_id, after=None, limit=None
):  # noqa: E501
    """List checkpoints for a fine-tuning job.

     # noqa: E501

    :param fine_tuning_job_id: The ID of the fine-tuning job to get checkpoints for.
    :type fine_tuning_job_id: str
    :param after: Identifier for the last checkpoint ID from the previous pagination request.
    :type after: str
    :param limit: Number of checkpoints to retrieve.
    :type limit: int

    :rtype: Union[ListFineTuningJobCheckpointsResponse, Tuple[ListFineTuningJobCheckpointsResponse, int], Tuple[ListFineTuningJobCheckpointsResponse, int, Dict[str, str]]
    """
    raise NotImplementedError


def list_paginated_fine_tuning_jobs(after=None, limit=None):  # noqa: E501
    """List your organization&#39;s fine-tuning jobs

     # noqa: E501

    :param after: Identifier for the last job from the previous pagination request.
    :type after: str
    :param limit: Number of fine-tuning jobs to retrieve.
    :type limit: int

    :rtype: Union[ListPaginatedFineTuningJobsResponse, Tuple[ListPaginatedFineTuningJobsResponse, int], Tuple[ListPaginatedFineTuningJobsResponse, int, Dict[str, str]]
    """
    raise NotImplementedError


# def retrieve_fine_tuning_job(fine_tuning_job_id):  # noqa: E501
def retrieve_fine_tuning_job(fine_tuning_job_id: str, token_info: dict, user: str):
    """Get info about a fine-tuning job.  [Learn more about fine-tuning](/docs/guides/fine-tuning)

     # noqa: E501

    :param fine_tuning_job_id: The ID of the fine-tuning job.
    :type fine_tuning_job_id: str

    :rtype: Union[FineTuningJob, Tuple[FineTuningJob, int], Tuple[FineTuningJob, int, Dict[str, str]]
    """
    # fine_tuning_job: FineTuningJob = instance_store._shared_instance_state["fine_tuning_jobs"].get(fine_tuning_job_id)
    fine_tuning_job: FineTuningJob = instance_store.select(
        FineTuningJob,
        # uuid.UUID(fine_tuning_job_id, version=4),
        id=fine_tuning_job_id,
    )

    return fine_tuning_job.model_dump(mode="json")



def create_image(body: dict, token_info: dict, user: str):
    """Creates an image given a prompt.

     # noqa: E501

    :param create_image_request:
    :type create_image_request: dict | bytes

    :rtype: Union[ImagesResponse, Tuple[ImagesResponse, int], Tuple[ImagesResponse, int, Dict[str, str]]
    """
    # if connexion.request.is_json:
    #     create_image_request = CreateImageRequest.from_dict(connexion.request.get_json())  # noqa: E501
    # raise NotImplementedErrora

    # print('args: ', args)
    # kwargs:  {'body': {'prompt': 'a white siamese cat', 'model': 'dall-e-3', 'n': 1, 'quality': 'standard', 'size': '1024x1024'}, 'user': 'user_id', 'token_info': {'uid': 'user_id'}}
    # print('kwargs: ', kwargs)

    # assert body.get("user") == user

    # width, height = map(int, body.get("size", "512x512").split("x"))

    # output: List[PIL.Image.Image] = stable_diffusion.txt_to_img(
    #     body.get("prompt"),
    #     height=height,
    #     # sample_steps=5, # default is 20
    #     width=width,
    # )

    # image: PIL.Image.Image = output[0]

    # # buffered = BytesIO()
    # # image.save(buffered, format="PNG")
    # # image.save(buffered, format="JSON")
    # # b64_json = str(base64.b64encode(buffered.getvalue()))

    # # s: ReadableBuffer
    # # base64.urlsafe_b64encode(s)
    # b64_json: str = base64.b64encode(image._repr_png_()).decode("utf-8")

    # images: List[Image] = [
    #     Image(
    #         b64_json=b64_json,
    #         revised_prompt=body.get("prompt"),
    #         url=None,
    #     )
    # ]

    # return ImagesResponse(
    #     created=int(time.time()),
    #     data=images,
    # ).model_dump(mode="json")

    raise NotImplementedError


def create_image_edit(
    image,
    prompt,
    mask=None,
    model=None,
    n=None,
    size=None,
    response_format=None,
    user=None,
):  # noqa: E501
    """Creates an edited or extended image given an original image and a prompt.

     # noqa: E501

    :param image: The image to edit. Must be a valid PNG file, less than 4MB, and square. If mask is not provided, image must have transparency, which will be used as the mask.
    :type image: str
    :param prompt: A text description of the desired image(s). The maximum length is 1000 characters.
    :type prompt: str
    :param mask: An additional image whose fully transparent areas (e.g. where alpha is zero) indicate where &#x60;image&#x60; should be edited. Must be a valid PNG file, less than 4MB, and have the same dimensions as &#x60;image&#x60;.
    :type mask: str
    :param model:
    :type model: dict | bytes
    :param n: The number of images to generate. Must be between 1 and 10.
    :type n: int
    :param size: The size of the generated images. Must be one of &#x60;256x256&#x60;, &#x60;512x512&#x60;, or &#x60;1024x1024&#x60;.
    :type size: str
    :param response_format: The format in which the generated images are returned. Must be one of &#x60;url&#x60; or &#x60;b64_json&#x60;. URLs are only valid for 60 minutes after the image has been generated.
    :type response_format: str
    :param user: A unique identifier representing your end-user, which can help OpenAI to monitor and detect abuse. [Learn more](/docs/guides/safety-best-practices/end-user-ids).
    :type user: str

    :rtype: Union[ImagesResponse, Tuple[ImagesResponse, int], Tuple[ImagesResponse, int, Dict[str, str]]
    """
    raise NotImplementedError


def create_image_variation(
    image, model=None, n=None, response_format=None, size=None, user=None
):  # noqa: E501
    """Creates a variation of a given image.

     # noqa: E501

    :param image: The image to use as the basis for the variation(s). Must be a valid PNG file, less than 4MB, and square.
    :type image: str
    :param model:
    :type model: dict | bytes
    :param n: The number of images to generate. Must be between 1 and 10. For &#x60;dall-e-3&#x60;, only &#x60;n&#x3D;1&#x60; is supported.
    :type n: int
    :param response_format: The format in which the generated images are returned. Must be one of &#x60;url&#x60; or &#x60;b64_json&#x60;. URLs are only valid for 60 minutes after the image has been generated.
    :type response_format: str
    :param size: The size of the generated images. Must be one of &#x60;256x256&#x60;, &#x60;512x512&#x60;, or &#x60;1024x1024&#x60;.
    :type size: str
    :param user: A unique identifier representing your end-user, which can help OpenAI to monitor and detect abuse. [Learn more](/docs/guides/safety-best-practices/end-user-ids).
    :type user: str

    :rtype: Union[ImagesResponse, Tuple[ImagesResponse, int], Tuple[ImagesResponse, int, Dict[str, str]]
    """
    raise NotImplementedError



from datetime import date, datetime  # noqa: F401
from typing import Dict, List  # noqa: F401

from timestep.api.ap.v1 import util
from timestep.api.ap.v1.models.artifact import Artifact  # noqa: E501
from timestep.api.ap.v1.models.base_model import Model
from timestep.api.ap.v1.models.pagination import Pagination  # noqa: E501


class TaskArtifactsListResponse(Model):
    """NOTE: This class is auto generated by OpenAPI Generator (https://openapi-generator.tech).

    Do not edit the class manually.
    """

    def __init__(self, artifacts=None, pagination=None):  # noqa: E501
        """TaskArtifactsListResponse - a model defined in OpenAPI

        :param artifacts: The artifacts of this TaskArtifactsListResponse.  # noqa: E501
        :type artifacts: List[Artifact]
        :param pagination: The pagination of this TaskArtifactsListResponse.  # noqa: E501
        :type pagination: Pagination
        """
        self.openapi_types = {"artifacts": List[Artifact], "pagination": Pagination}

        self.attribute_map = {"artifacts": "artifacts", "pagination": "pagination"}

        self._artifacts = artifacts
        self._pagination = pagination

    @classmethod
    def from_dict(cls, dikt) -> "TaskArtifactsListResponse":
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The TaskArtifactsListResponse of this TaskArtifactsListResponse.  # noqa: E501
        :rtype: TaskArtifactsListResponse
        """
        return util.deserialize_model(dikt, cls)

    @property
    def artifacts(self) -> List[Artifact]:
        """Gets the artifacts of this TaskArtifactsListResponse.


        :return: The artifacts of this TaskArtifactsListResponse.
        :rtype: List[Artifact]
        """
        return self._artifacts

    @artifacts.setter
    def artifacts(self, artifacts: List[Artifact]):
        """Sets the artifacts of this TaskArtifactsListResponse.


        :param artifacts: The artifacts of this TaskArtifactsListResponse.
        :type artifacts: List[Artifact]
        """
        if artifacts is None:
            raise ValueError(
                "Invalid value for `artifacts`, must not be `None`"
            )  # noqa: E501

        self._artifacts = artifacts

    @property
    def pagination(self) -> Pagination:
        """Gets the pagination of this TaskArtifactsListResponse.


        :return: The pagination of this TaskArtifactsListResponse.
        :rtype: Pagination
        """
        return self._pagination

    @pagination.setter
    def pagination(self, pagination: Pagination):
        """Sets the pagination of this TaskArtifactsListResponse.


        :param pagination: The pagination of this TaskArtifactsListResponse.
        :type pagination: Pagination
        """
        if pagination is None:
            raise ValueError(
                "Invalid value for `pagination`, must not be `None`"
            )  # noqa: E501

        self._pagination = pagination



from datetime import date, datetime  # noqa: F401
from typing import Dict, List  # noqa: F401

from timestep.api.ap.v1 import util
from timestep.api.ap.v1.models.base_model import Model


class TaskRequestBody(Model):
    """NOTE: This class is auto generated by OpenAPI Generator (https://openapi-generator.tech).

    Do not edit the class manually.
    """

    def __init__(self, input=None, additional_input=None):  # noqa: E501
        """TaskRequestBody - a model defined in OpenAPI

        :param input: The input of this TaskRequestBody.  # noqa: E501
        :type input: str
        :param additional_input: The additional_input of this TaskRequestBody.  # noqa: E501
        :type additional_input: object
        """
        self.openapi_types = {"input": str, "additional_input": object}

        self.attribute_map = {"input": "input", "additional_input": "additional_input"}

        self._input = input
        self._additional_input = additional_input

    @classmethod
    def from_dict(cls, dikt) -> "TaskRequestBody":
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The TaskRequestBody of this TaskRequestBody.  # noqa: E501
        :rtype: TaskRequestBody
        """
        return util.deserialize_model(dikt, cls)

    @property
    def input(self) -> str:
        """Gets the input of this TaskRequestBody.

        Input prompt for the task.  # noqa: E501

        :return: The input of this TaskRequestBody.
        :rtype: str
        """
        return self._input

    @input.setter
    def input(self, input: str):
        """Sets the input of this TaskRequestBody.

        Input prompt for the task.  # noqa: E501

        :param input: The input of this TaskRequestBody.
        :type input: str
        """

        self._input = input

    @property
    def additional_input(self) -> object:
        """Gets the additional_input of this TaskRequestBody.

        Input parameters for the task. Any value is allowed.  # noqa: E501

        :return: The additional_input of this TaskRequestBody.
        :rtype: object
        """
        return self._additional_input

    @additional_input.setter
    def additional_input(self, additional_input: object):
        """Sets the additional_input of this TaskRequestBody.

        Input parameters for the task. Any value is allowed.  # noqa: E501

        :param additional_input: The additional_input of this TaskRequestBody.
        :type additional_input: object
        """

        self._additional_input = additional_input



def create_embedding(create_embedding_request):  # noqa: E501
    """Creates an embedding vector representing the input text.

     # noqa: E501

    :param create_embedding_request:
    :type create_embedding_request: dict | bytes

    :rtype: Union[CreateEmbeddingResponse, Tuple[CreateEmbeddingResponse, int], Tuple[CreateEmbeddingResponse, int, Dict[str, str]]
    """
    raise NotImplementedError



import inspect
import os

import typer
from sqlmodel import SQLModel, create_engine

from timestep.server import main as timestep_serve
from timestep.worker import main as timestep_train

# TODO: move these to the config/env
app_dir = typer.get_app_dir(__package__)
default_llamafile_filename = "TinyLlama-1.1B-Chat-v1.0.F16.llamafile"
default_llamafile_url = f"https://huggingface.co/Mozilla/TinyLlama-1.1B-Chat-v1.0-llamafile/resolve/main/{default_llamafile_filename}?download=true"

os.makedirs(f"{app_dir}/data", exist_ok=True)
os.makedirs(f"{app_dir}/models", exist_ok=True)
os.makedirs(f"{app_dir}/work", exist_ok=True)

engine = create_engine(f"sqlite:///{app_dir}/database.db")
SQLModel.metadata.create_all(
    bind=engine,
    checkfirst=True,
    tables=None,
)


def get_help_message():
    is_readme_context = inspect.getmodule(inspect.stack()[1][0]) is None

    return """
Timestep AI CLI - free, local-first, open-source AI
""" + (
        """
**Setup**:

### Development

```console
$ python3 -m pip install --upgrade pip
$ python3 -m pip install --user pipx
$ python3 -m pipx ensurepath
$ pipx install poetry==1.8.3
$ cp .env.example .env
$ direnv allow # See https://direnv.net/#getting-started to install direnv on your platform
$ make
```

### Library

```console
$ python3 -m pip install --upgrade pip
$ python3 -m pip install --user pipx
$ python3 -m pipx ensurepath
$ pipx install timestep
```

**Pre-requisites**:

```console
$ prefect server start
$ prefect worker start --pool "default"
```
"""
        if is_readme_context
        else ""
    )


typer_app = typer.Typer(
    help=get_help_message(),
    no_args_is_help=True,
)

# typer_app.add_typer(llamafile_typer_app, name="up")


@typer_app.callback()
def main():
    f"""
    Timestep AI CLI
    """


@typer_app.command()
def evals():
    """
    Run evaluations.
    """
    typer.echo("Running evaluations...")

    raise NotImplementedError


@typer_app.command()
def serve(
    # host="0.0.0.0",
    # llamafile_path=f"./models/{default_llamafile_filename}", # TODO: namespace under llamafile, include port, etc.
    # port=8080,
):
    """
    Run serving.
    """
    typer.echo("Running serving...")

    timestep_serve(
        # host=host,
        # llamafile_path=llamafile_path,
        # port=port,
    )


@typer_app.command()
def train():
    """
    Run training.
    """
    typer.echo("Running training...")

    timestep_train()


if __name__ == "__main__":
    typer_app()



import os
from typing import Optional

# import controlflow as cf
import httpx
import openai
import typer

# from langchain_openai import ChatOpenAI
from openai.types.beta.assistant import Assistant
from openai.types.beta.thread import Thread
from openai.types.beta.threads.run import Run
from openai.types.chat.chat_completion import ChatCompletion
from openai.types.file_object import FileObject
from openai.types.fine_tuning.fine_tuning_job import FineTuningJob, Hyperparameters
from prefect import flow, get_run_logger
from prefect_shell import ShellOperation

# from prefect_sqlalchemy import AsyncDriver, ConnectionComponents, SqlAlchemyConnector


app_dir = typer.get_app_dir("timestep")
# block_name = "timestep-ai-sql-alchemy-connector-block"
# connector = SqlAlchemyConnector(
#     connection_info=ConnectionComponents(
#         driver=AsyncDriver.SQLITE_AIOSQLITE, database=f"{app_dir}/database.db"
#     )
# )

# cf.default_model = ChatOpenAI(
#     api_key=os.environ.get("OPENAI_API_KEY"),
#     base_url=os.environ.get("OPENAI_BASE_URL"),
#     temperature=0.0,
# )


@flow
async def agent_flow(inputs: dict, work_type: str):
    logger = get_run_logger()
    logger.info(f"inputs: {inputs}")

    work_type = inputs.get("work_type", "train_agent")

    # await connector.save(block_name, overwrite=True)

    if work_type == "run_agent":
        await run_agent_flow(run_id=inputs.get("run_id"))

    elif work_type == "train_agent":
        await train_agent_flow(
            fine_tuning_job_id=inputs.get("fine_tuning_job_id"),
            suffix=inputs.get("suffix"),
        )

    else:
        raise NotImplementedError


@flow(log_prints=True)
async def run_agent_flow(run_id: str):
    # run: Run = Run(**await get_run(run_id=run_id, token_info=token_info, thread_id=thread_id, user=user))
    run: Run = await get_run(
        run_id=run_id, token_info=token_info, thread_id=thread_id, user=user
    )
    assistant: Assistant = await get_assistant(
        assistant_id=run.assistant_id, token_info=token_info, user=user
    )
    thread: Thread = await get_thread(
        token_info=token_info, thread_id=thread_id, user=user
    )

    messages = await list_messages(
        limit=-1,
        order="asc",
        token_info=token_info,
        thread_id=thread_id,
        user=user,
    )

    system_message = {
        "content": "You are a helpful assistant with access to tools. Please be honest and do not hallucinate.",
        "role": "system",
    }

    messages = [system_message] + messages
    print("messages: ", messages)

    def get_message_content(message):
        content = message.get("content")

        if type(content) == list:
            if message.get("role") == "user":
                return [
                    {
                        "text": _content.get("text").get("value"),
                        "type": _content.get("type"),
                    }
                    for _content in content
                ]

            else:
                assert len(content) == 1
                return content[0].get("text").get("value")

        else:
            return content

    body = {  # CompletionCreateParamsNonStreaming
        "messages": [
            {
                "content": get_message_content(message),
                "role": message.get("role"),
            }
            for message in messages
        ],
        "model": assistant.model,
        "tools": assistant.tools,
    }
    print("body: ", body)

    # TODO: use service instead?
    chat_completion: ChatCompletion = ChatCompletion(
        **await create_chat_completion(body=body, token_info=token_info, user=user)
    )  # TODO: handle streaming use case
    print("chat_completion: ", chat_completion)

    choice = chat_completion.choices[0]
    print("choice: ", choice)

    finish_reason = choice.finish_reason
    print(
        "finish_reason: ", finish_reason
    )  # "stop", "length", "tool_calls", "content_filter", "function_call"

    response_message = chat_completion.choices[0].message

    tool_calls = response_message.tool_calls

    if tool_calls:
        raise NotImplementedError

    await create_message(
        body={
            "content": response_message.content,
            "role": response_message.role,
        },
        token_info=token_info,
        thread_id=thread_id,
        user=user,
    )

    # time.sleep(1)  # wait for the process to finish

    modify_run_request = {
        "status": "completed",
    }

    await modify_run(
        modify_run_request=modify_run_request,
        run_id=run_id,
        token_info=token_info,
        thread_id=thread_id,
        user=user,
    )

    # run: Run = Run(**await get_run(run_id=run_id, token_info=token_info, thread_id=thread_id, user=user))

    # return run


# @task
# async def select_fine_tuning_job_by_id(id: str):
#     async with await SqlAlchemyConnector.load(block_name) as connector:
#         # results = await connector.fetch_one(f"SELECT * FROM fine_tuning_jobs WHERE id = '{unhyphenated_id}';") # TODO: use prepared statement equivalent
#         results = await connector.fetch_one(
#             f"""SELECT
#                     created_at,
#                     error,
#                     estimated_finish,
#                     fine_tuned_model,
#                     finished_at,
#                     hyperparameters,
#                     integrations,
#                     model,
#                     object,
#                     organization_id,
#                     result_files,
#                     seed,
#                     status,
#                     trained_tokens,
#                     training_file,
#                     validation_file
#                 FROM fine_tuning_jobs
#                 WHERE id = :id;""",
#             parameters={"id": str(uuid.UUID(id, version=4)).replace("-", "")},
#         )
#         print("results: ", results)

#         return FineTuningJob(
#             id=id,
#             created_at=results[0],
#             error=json.loads(results[1]),
#             estimated_finish=results[2],
#             fine_tuned_model=results[3],
#             finished_at=results[4],
#             hyperparameters=json.loads(results[5]),
#             integrations=json.loads(results[6]),
#             model=results[7],
#             object=results[8],
#             organization_id=results[9],
#             result_files=json.loads(results[10]),
#             seed=results[11],
#             status=results[12],
#             trained_tokens=results[13],
#             training_file=results[14],
#             validation_file=results[15],
#         )
#         # return results

#     # return all_rows


@flow(log_prints=True)
async def train_agent_flow(fine_tuning_job_id: str, suffix: Optional[str] = None):
    logger = get_run_logger()
    logger.info("INFO level log message.")

    print("=== training model ===")
    print("fine_tuning_job_id: ", fine_tuning_job_id)

    working_dir = f"{app_dir}/work/{fine_tuning_job_id}"
    os.makedirs(working_dir, exist_ok=True)

    # fine_tuning_job: FineTuningJob = instance_store._shared_instance_state["fine_tuning_jobs"].get(fine_tuning_job_id)
    # fine_tuning_job: FineTuningJob = await select_fine_tuning_job_by_id(
    #     fine_tuning_job_id
    # )

    openai_client = openai.OpenAI(
        api_key=os.environ.get("OPENAI_API_KEY"),
        base_url=os.environ.get("OPENAI_BASE_URL"),
    )

    fine_tuning_job: FineTuningJob = (
        openai_client.fine_tuning.jobs.retrieve(  # TODO: is there an async version?
            fine_tuning_job_id=fine_tuning_job_id,
        )
    )

    print("fine_tuning_job: ", fine_tuning_job)

    hyperparameters: Hyperparameters = fine_tuning_job.hyperparameters
    max_epochs = -1 if hyperparameters.n_epochs == "auto" else hyperparameters.n_epochs

    # model: Llama = instance_store._shared_instance_state["models"][fine_tuning_job.model]
    async with httpx.AsyncClient() as client:
        response = await client.get(
            f'{os.environ.get("OPENAI_BASE_URL").split("/api")[0]}/v2/models/{fine_tuning_job.model}'
        )
        print("response: ", response)
        model_info = response.json()
        print("model_info: ", model_info)

    model_path: str = model_info.get("model_path")
    print("model_path: ", model_path)

    lora_base: str = model_info.get("lora_base")
    print("lora_base: ", lora_base)

    lora_path: str = model_info.get("lora_path")
    print("lora_path: ", lora_path)

    lora_scale: str = model_info.get("lora_scale")
    print("lora_scale: ", lora_scale)

    # fine_tuning_job = FineTuningJob(
    #     id=str(uuid.uuid4()),
    #     created_at=int(time.time()),
    #     hyperparameters=hyperparameters,
    #     model=body.get("model"),
    #     object="fine_tuning.job",
    #     organization_id="organization_id",
    #     result_files=[],
    #     seed=body.get("seed", 42),
    #     status="queued", # Literal['validating_files', 'queued', 'running', 'succeeded', 'failed', 'cancelled'],
    #     training_file=body.get("training_file"),
    #     validation_file=body.get("validation_file"),
    # )

    # fine_tuning_jobs_timestep.db[fine_tuning_job.id] = fine_tuning_job

    # fine_tuning_job.__setattr__(
    #     "status",
    #     "running",
    # )

    # fine_tuning_job_event = FineTuningJobEvent(
    #     id=str(uuid.uuid4()),
    #     created_at=int(time.time()),
    #     level="info",
    #     message=f"The job has begun",
    #     object="fine_tuning.job.event",
    # )

    # # train_data = f"{os.getcwd()}/notebooks/openai-cookbook/examples/tmp_recipe_finetune_training.jsonl"
    # # train_data = fine_tuning_job.training_file
    training_file_id = fine_tuning_job.training_file
    validation_file_id = fine_tuning_job.validation_file

    # print('training_file_id: ', training_file_id)
    # print('validation_file_id: ', validation_file_id)

    # training_file_object = instance_store._shared_instance_state["file_objects"][training_file_id]
    # validation_file_object = instance_store._shared_instance_state["file_objects"][validation_file_id]

    training_file_object: FileObject = openai_client.files.retrieve(
        file_id=training_file_id
    )

    print("training_file_object: ", training_file_object)

    if validation_file_id:
        validation_file_object: FileObject = openai_client.files.retrieve(
            file_id=validation_file_id
        )

        print("validation_file_object: ", validation_file_object)

    # str(uuid.UUID(id, version=4)).replace("-", "")

    # for long running operations, you can use a context manager
    async with ShellOperation(
        commands=[
            # TODO: https://github.com/ggerganov/llama.cpp/blob/master/docs/backend/SYCL.md
            f"git clone https://github.com/ggerganov/llama.cpp {app_dir}/3rdparty/llama.cpp",
            f"make -C {app_dir}/3rdparty/llama.cpp",  # TODO: handle windows, termux, etc.
            f"""{app_dir}/3rdparty/llama.cpp/llama-finetune \
--adam-iter 30 \
--batch 8 \
--checkpoint-in "chk-lora-{fine_tuning_job.model}-{fine_tuning_job.id}-LATEST.gguf" \
--checkpoint-out "chk-lora-{fine_tuning_job.model}-{fine_tuning_job.id}-ITERATION.gguf" \
--ctx 64 \
--epochs {max_epochs} \
--lora-out "lora-{fine_tuning_job.model}-{fine_tuning_job.id}-ITERATION.bin"  \
--model-base {model_path} \
--n-gpu-layers 0 \
--save-every 10 \
--seed {fine_tuning_job.seed} \
--threads 6 \
--train-data "{app_dir}/data/{training_file_object.id}/{training_file_object.filename}" \
--use-checkpointing
            """,
        ],
        env={
            "CMAKE_ARGS": os.getenv("CMAKE_ARGS", ""),
        },
        working_dir=working_dir,
    ) as llama_finetune_operation:

        # trigger runs the process in the background
        llama_finetune_process = await llama_finetune_operation.trigger()

        # then do other things here in the meantime, like download another file
        ...

        # when you're ready, wait for the process to finish
        await llama_finetune_process.wait_for_completion()

        # if you'd like to get the output lines, you can use the `fetch_result` method
        output_lines = await llama_finetune_process.fetch_result()
        print("output_lines: ", output_lines)

    #     for output_line in output_lines:
    #         print('output_line: ', output_line)

    #         fine_tuning_job_event = FineTuningJobEvent(
    #             id=str(uuid.uuid4()),
    #             created_at=int(time.time()),
    #             level="info",
    #             message=output_line,
    #             object="fine_tuning.job.event",
    #         )

    #         instance_store._shared_instance_state["fine_tuning_job_events"][fine_tuning_job.id].append(fine_tuning_job_event)

    #     fine_tuning_job.__setattr__(
    #         "fine_tuned_model",
    #         # "ft:gpt-3.5-turbo-0613:personal:recipe-ner:8PjmcwDH"
    #         f"ft:{fine_tuning_job.model}:{fine_tuning_job.organization_id}:{suffix}:{fine_tuning_job.id}"
    #     )

    #     fine_tuning_job_event = FineTuningJobEvent(
    #         id=str(uuid.uuid4()),
    #         created_at=int(time.time()),
    #         level="info",
    #         message=f"New fine-tuned model created: {fine_tuning_job.fine_tuned_model}",
    #         object="fine_tuning.job.event",
    #     )

    #     instance_store._shared_instance_state["fine_tuning_job_events"][fine_tuning_job.id].append(fine_tuning_job_event)

    #     fine_tuning_job.__setattr__(
    #         "status",
    #         "succeeded",
    #     )

    #     fine_tuning_job_event = FineTuningJobEvent(
    #         id=str(uuid.uuid4()),
    #         created_at=int(time.time()),
    #         level="info",
    #         message="The job has successfully completed",
    #         object="fine_tuning.job.event",
    #     )

    #     instance_store._shared_instance_state["fine_tuning_job_events"][fine_tuning_job.id].append(fine_tuning_job_event)

    #     base_model: Llama = instance_store._shared_instance_state["models"][fine_tuning_job.model]

    #     # model = Llama(
    #     #     chat_format=base_model.chat_format,
    #     #     lora_base=f"{os.getcwd()}/{base_model.model_path}",
    #     #     lora_path=f"{os.getcwd()}/{working_dir}/lora-{fine_tuning_job.model}-{task}-ITERATION.bin",
    #     #     model_path=base_model.model_path,
    #     #     n_ctx=base_model.n_ctx(),
    #     # )

    #     # model = serve_model()

    #     # instance_store._shared_instance_state["models"][fine_tuning_job.fine_tuned_model] = model


# async def main():
# def main(work_type="train_agent"):
def main(training_file_id):
    # flow_run: FlowRun = run_deployment(
    #     # idempotency_key=fine_tuning_job.id,
    #     name="agent-flow/agent-flow-deployment",
    #     # parameters={"names": ["Alice", "Bob"]},
    #     parameters={
    #         "inputs": {
    #             "fine_tuning_job_id": "607946aa-ecd8-4e06-8163-9ca70d754238",
    #             "suffix": "shakespeare",
    #         },
    #         "work_type": work_type,
    #     },
    #     # job_variables={"env": {"MY_ENV_VAR": "staging"}},
    #     timeout=0,
    # )

    openai_client = openai.OpenAI(
        api_key=os.environ.get("OPENAI_API_KEY"),
        base_url=os.environ.get("OPENAI_BASE_URL"),
    )

    # print("flow_run: ", flow_run)
    fine_tuning_job: FineTuningJob = openai_client.fine_tuning.jobs.create(
        model="OpenLLaMA-3Bv2",
        suffix="shakespeare",
        training_file=training_file_id,
    )

    # job_id = fine_tuning_job.id


if __name__ == "__main__":
    # main(work_type="train_agent")
    main()



# def create_speech(create_speech_request):  # noqa: E501
def create_speech(*args, **kwargs):
    """Generates audio from the input text.

    :param create_speech_request:
    :type create_speech_request: dict | bytes

    :rtype: Union[file, Tuple[file, int], Tuple[file, int, Dict[str, str]]
    """
    raise NotImplementedError


def create_transcription(
    file,
    model,
    language=None,
    prompt=None,
    response_format=None,
    temperature=None,
    timestamp_granularities=None,
):  # noqa: E501
    """Transcribes audio into the input language.

     # noqa: E501

    :param file: The audio file object (not file name) to transcribe, in one of these formats: flac, mp3, mp4, mpeg, mpga, m4a, ogg, wav, or webm.
    :type file: str
    :param model:
    :type model: dict | bytes
    :param language: The language of the input audio. Supplying the input language in [ISO-639-1](https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes) format will improve accuracy and latency.
    :type language: str
    :param prompt: An optional text to guide the model&#39;s style or continue a previous audio segment. The [prompt](/docs/guides/speech-to-text/prompting) should match the audio language.
    :type prompt: str
    :param response_format: The format of the transcript output, in one of these options: &#x60;json&#x60;, &#x60;text&#x60;, &#x60;srt&#x60;, &#x60;verbose_json&#x60;, or &#x60;vtt&#x60;.
    :type response_format: str
    :param temperature: The sampling temperature, between 0 and 1. Higher values like 0.8 will make the output more random, while lower values like 0.2 will make it more focused and deterministic. If set to 0, the model will use [log probability](https://en.wikipedia.org/wiki/Log_probability) to automatically increase the temperature until certain thresholds are hit.
    :type temperature:
    :param timestamp_granularities: The timestamp granularities to populate for this transcription. &#x60;response_format&#x60; must be set &#x60;verbose_json&#x60; to use timestamp granularities. Either or both of these options are supported: &#x60;word&#x60;, or &#x60;segment&#x60;. Note: There is no additional latency for segment timestamps, but generating word timestamps incurs additional latency.
    :type timestamp_granularities: List[str]

    :rtype: Union[CreateTranscription200Response, Tuple[CreateTranscription200Response, int], Tuple[CreateTranscription200Response, int, Dict[str, str]]
    """
    raise NotImplementedError


def create_translation(
    file, model, prompt=None, response_format=None, temperature=None
):  # noqa: E501
    """Translates audio into English.

     # noqa: E501

    :param file: The audio file object (not file name) translate, in one of these formats: flac, mp3, mp4, mpeg, mpga, m4a, ogg, wav, or webm.
    :type file: str
    :param model:
    :type model: dict | bytes
    :param prompt: An optional text to guide the model&#39;s style or continue a previous audio segment. The [prompt](/docs/guides/speech-to-text/prompting) should be in English.
    :type prompt: str
    :param response_format: The format of the transcript output, in one of these options: &#x60;json&#x60;, &#x60;text&#x60;, &#x60;srt&#x60;, &#x60;verbose_json&#x60;, or &#x60;vtt&#x60;.
    :type response_format: str
    :param temperature: The sampling temperature, between 0 and 1. Higher values like 0.8 will make the output more random, while lower values like 0.2 will make it more focused and deterministic. If set to 0, the model will use [log probability](https://en.wikipedia.org/wiki/Log_probability) to automatically increase the temperature until certain thresholds are hit.
    :type temperature:

    :rtype: Union[CreateTranslation200Response, Tuple[CreateTranslation200Response, int], Tuple[CreateTranslation200Response, int, Dict[str, str]]
    """
    raise NotImplementedError



def delete_model(model):  # noqa: E501
    """Delete a fine-tuned model. You must have the Owner role in your organization to delete a model.

     # noqa: E501

    :param model: The model to delete
    :type model: str

    :rtype: Union[DeleteModelResponse, Tuple[DeleteModelResponse, int], Tuple[DeleteModelResponse, int, Dict[str, str]]
    """
    raise NotImplementedError


def list_models():  # noqa: E501
    """Lists the currently available models, and provides basic information about each one such as the owner and availability.

     # noqa: E501


    :rtype: Union[ListModelsResponse, Tuple[ListModelsResponse, int], Tuple[ListModelsResponse, int, Dict[str, str]]
    """
    raise NotImplementedError


def retrieve_model(model: str, token_info: dict, user: str):
    """Retrieves a model instance, providing basic information about the model such as the owner and permissioning.

     # noqa: E501

    :param model: The ID of the model to use for this request
    :type model: str

    :rtype: Union[Model, Tuple[Model, int], Tuple[Model, int, Dict[str, str]]
    """
    # print('args: ', args)
    # print('kwargs: ', kwargs)

    # model_info = model_service.retrieve_model(model_id=model)

    # print("model_info: ", model_info)

    # return Model(
    #     id=model,
    #     created=0,
    #     object="model",
    #     owned_by=user,
    # ).model_dump(mode="json")

    raise NotImplementedError



def info_from_ApiKeyAuth(token):
    """
    Check and retrieve authentication information from custom bearer token.
    Returned value will be passed in 'token_info' parameter of your operation function, if there is one.
    'sub' or 'uid' will be set in 'user' parameter of your operation function, if there is one.

    :param token Token provided by Authorization header
    :type token: str
    :return: Decoded token information or None if token is invalid
    :rtype: dict | None
    """
    return {"uid": "user_id"}
